"""Resilience tests."""
from tests import base as b
from tests.test_interface import Perforator

import sqlalchemy as sa


@b.add_memory()
class CleanData(b.TestCase):
    """Testcase that cleans __meta_dataclasses__ table while membank has been loaded."""

    def test_delete_meta_contents(self):
        """Clean __meta_dataclasses__ table."""
        p = Perforator("test")
        self.memory.put(p)
        self.assertTrue(self.memory.get.perforator(name="test"))
        self.commit_stmt("DELETE FROM __meta_dataclasses__")
        result = self.memory.get.perforator(name="test")
        self.assertIsNotNone(result)
        self.assertTrue(result.name, "test")
        self.memory.put(Perforator("some other perforator"))
        self.memory.put(Perforator("some more perforators"))
        self.assertTrue(self.memory.get.perforator(name="some other perforator"))
        self.assertEqual(p, self.memory.get.perforator(name="test"))

    def test_delete_meta_dataclasses(self):
        """Drop __meta_dataclasses__ table."""
        p = Perforator("test")
        self.memory.put(p)
        self.commit_stmt("DROP TABLE __meta_dataclasses__")
        result = self.memory.get(self.memory.perforator.name == "test")
        self.assertIsNotNone(result)
        self.assertIsNotNone(self.memory.get.perforator(name="test"))
        self.memory.put(p)

    def commit_stmt(self, stmt):
        """Destroy meta_dataclasses table."""
        engine = self.memory._get_engine()
        with engine.connect() as conn:
            conn.execute(sa.text(stmt))
            conn.commit()



"""Module supports dataclass storage and retrieval."""

import pickle
from dataclasses import dataclass, make_dataclass

from membank import errors as e
from membank.datamethods import (create_table, get_item,
                                 introspect_table_fields, update_item)


@dataclass
class TableClass:
    """Maps a dataclass to a Table."""

    table: str = ""
    classload: bytes = b""


class Mapper:
    """Interface to store and retrieve dataclasses."""

    def __init__(self, engine, metadata):
        """Initialise."""
        self.engine = engine
        self.metadata = metadata
        if "__meta_dataclasses__" not in metadata:
            self._create_meta()
        self.sql_table = self.metadata.tables["__meta_dataclasses__"]

    def _create_meta(self):
        """Create __meta_dataclasses__ table."""
        create_table("__meta_dataclasses__", TableClass(), self.engine)
        self.metadata.reflect(bind=self.engine)

    def _get_temporary_fallback_class(self, table):
        """Return a temporary fallback dataclass for given table."""
        self.metadata.reflect(bind=self.engine)
        cls_name = table.capitalize() + "Fallback"
        fields = introspect_table_fields(self.metadata.tables[table])
        return make_dataclass(cls_name, fields)

    def _handle_missing_meta(self, func, *args, **kwargs):
        """Run function such that missing __meta_dataclasses are taken into account.

        Upon exception that table __meta_dataclasses__ is missing, fix it and reraise,
        otherwise just raises error.
        """
        try:
            return func(*args, **kwargs)
        except e.MemoryTableDoesNotExist:
            self._create_meta()
            raise

    def get_class(self, table):
        """Return dataclass representing table."""
        try:
            table_class = self._handle_missing_meta(
                get_item, self.sql_table, self.engine, TableClass, **{"table": table}
            )
        except e.MemoryTableDoesNotExist:
            table_class = None
        if table_class:
            return pickle.loads(table_class.classload)
        return self._get_temporary_fallback_class(table)

    def put_class(self, table, table_class):
        """Store dataclass representing table."""
        classload = pickle.dumps(table_class)
        args = (
            self.sql_table,
            self.engine,
            TableClass(table=table, classload=classload),
        )
        kargs = {"key": "table"}
        try:
            self._handle_missing_meta(update_item, *args, **kargs)
        except e.MemoryTableDoesNotExist:
            update_item(*args, **kargs)



"""Connection utlity for SQLAlchemy and asyncpg."""

import json
from collections import namedtuple

import asyncpg
import sqlalchemy as sa


async def asyncpg_custom_types(conn):
    """Return custom types for asyncpg."""
    await conn.set_type_codec(
        "json",
        encoder=lambda x: x,
        decoder=json.loads,
        schema="pg_catalog",
    )
    await conn.set_type_codec(
        "jsonb",
        encoder=lambda x: x,
        decoder=json.loads,
        schema="pg_catalog",
    )


# pylint: disable=too-few-public-methods
class NamedRecord(asyncpg.Record):
    """Record class that supports namedtuple like behaviour."""

    def __getattr__(self, key):
        """Return value for key."""
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key) from None


class SQLAlchemyResult:
    """Result class that supports some of the SQLAlchemy result methods.

    Object is returned from asyncpg execution, e.g session.execute
    """

    def __init__(self, result):
        """Load object with result from asyncpg execution."""
        self._mapping = result

    def first(self):
        """Return first row."""
        return self._mapping[0] if self._mapping else None

    def __iter__(self):
        """Return iterator."""
        return iter(self._mapping)

    def rowcount(self):
        """Return row count."""
        return len(self._mapping)


class SQLAlchemyDBAPI:
    """SQlAlchemy DBAPI silencer."""

    def cursor(self):
        """Return nothing."""


class SQLAlchemySupportedConnection(asyncpg.connection.Connection):
    """Connection class that supports SQLAlchemy stmt objects to be executed."""

    dialect = sa.dialects.postgresql.asyncpg.PGDialect_asyncpg()

    async def execute(self, query, *args, timeout=300):
        """Execute statement.

        If query appears to be SQLAlchemy stmt object, convert it to string.
        If query appears to be SQLAlchemy ORM object, convert it to valid stmt.
        """
        if isinstance(query, sa.sql.base.Executable):
            compiled_query = query.compile(
                dialect=sa.dialects.postgresql.asyncpg.PGDialect_asyncpg()
            )
            ctx = self.dialect.execution_ctx_cls._init_compiled(
                self.dialect,
                self,
                dbapi_connection=SQLAlchemyDBAPI(),
                execution_options={},
                compiled=compiled_query,
                parameters=tuple(),
                invoked_statement=None,
                extracted_parameters={},
            )
            stmt = ctx.statement
            params = ctx.parameters[0]
            if isinstance(query, sa.sql.expression.Delete):
                result = await super().execute(stmt, *params, timeout=timeout)
                SQLDelete = namedtuple("SQLDelete", ["rowcount"])
                rowcount = int(result.replace("DELETE ", ""))
                return SQLDelete(rowcount)
            return SQLAlchemyResult(await super().fetch(stmt, *params, timeout=timeout))
        return await super().execute(query, *args, timeout=timeout)



"""Exceptions for membank."""


class GeneralMemoryError(Exception):
    """All general errors in memory interface."""


class MemoryTableDoesNotExist(Exception):
    """Table does not exist in memory."""

    def __init__(self, table_name):
        """Initialise message."""
        msg = f"Table '{table_name}' does not exist"
        Exception.__init__(self, msg)


class MemoryFilteringError(Exception):
    """Error with filtering different tables at the same time."""

    def __init__(self, table1, table2):
        """Initialise message."""
        msg = f"Not possible to filter {table1} and {table2}"
        msg += " at the same time"
        Exception.__init__(self, msg)


class MemoryOutOfSyncError(Exception):
    """Error when memory is out of sync with database."""

    def __init__(self, table_name, name):
        """Initialise message."""
        msg = f"Object '{table_name}' appears to be out of sync"
        msg += f" as '{name}' is not found"
        Exception.__init__(self, msg)



"""Tests for membank.interface main API for library."""

import base64
import dataclasses
from dataclasses import dataclass
import datetime

import membank
from tests import base as b


def encode_team(team):
    """Encode dog in json."""
    return [encode_dog(i) for i in team]


def encode_dog(dog):
    """Encode dog in json."""
    res = dataclasses.asdict(dog)
    res["picture"] = base64.b64encode(res["picture"]).decode("utf-8")
    return res


def decode_dog(kwargs):
    """Decode dog from json."""
    kwargs["picture"] = base64.b64decode(kwargs["picture"].encode("utf-8"))
    return Dog(**kwargs)


# Test tables
@dataclass
class Dog:
    """Simple example from README."""

    breed: str
    color: str = "black"
    weight: float = 0.0
    data: dict = dataclasses.field(default_factory=dict)
    picture: bytes = b""
    alive: bool = True
    aliases: list = dataclasses.field(default_factory=list)


@dataclass
class UpdatedDog:
    """Dog that is updated."""

    breed: str
    color: str = "black"


@dataclass
class DogTeam:
    """Team of dogs."""

    name: str = dataclasses.field(metadata={"key": True})
    team: list = dataclasses.field(
        default_factory=list, metadata={"encode": encode_team}
    )

    def __post_init__(self):
        """Convert dictionaries in team list to dogs."""
        for i, val in enumerate(self.team):
            if not isinstance(val, Dog):
                self.team[i] = decode_dog(val)


@dataclass
class Cat:
    """Cat example."""

    id: str = dataclasses.field(default=None, metadata={"key": True})
    name: str = "Rudolf"
    color: str = "black"


@dataclass
class Transaction:
    """Example with pre post handling."""

    amount: float
    description: str
    timestamp: datetime.datetime = None
    id: str = dataclasses.field(default=None, metadata={"key": True})

    def __post_init__(self):
        """Add unique id to transaction."""
        if not self.timestamp:
            self.timestamp = datetime.datetime.now()
        if not self.id:
            self.id = f"special_id:{self.description}"


@dataclass
class Perforator:
    """Example with name attibute."""

    name: str


@b.add_memory()
class CleanData(b.TestCase):
    """Testcase on clean_all_data function in Memory."""

    def test(self):
        """clean_all_data wipes data but not tables."""
        self.memory.put(Dog("Puli"))
        self.memory.clean_all_data()
        self.memory.get("dog")
        self.memory.put(Dog("Puli"))


@b.add_memory(b.DBPath.RELATIVE, reset=True)
class Operator(b.TestCase):
    """Testcases on comparison operators."""

    def test_equal(self):
        """Equality on name."""
        self.memory.put(Perforator("perforate"))
        self.assertTrue(self.memory.get(self.memory.perforator.name == "perforate"))


@b.add_memory(b.DBPath.RELATIVE, reset=True)
class Delete(b.TestCase):
    """Delete a memory item."""

    def test_delete(self):
        """Delete an item."""
        booking = Transaction(50, "delete transaction")
        self.memory.put(booking)
        self.assertTrue(self.memory.get.transaction(id=booking.id))
        self.memory.delete(booking)
        self.assertFalse(self.memory.get.transaction(id=booking.id))


@b.add_memory(b.DBPath.RELATIVE, reset=True)
class GetList(b.TestCase):
    """Testcase on getting list of items instead of single."""

    def setUp(self):
        """Add transactions as testdata."""
        super().setUp()
        for i in range(10):
            booking = Transaction(50 + i, f"list transaction {i}")
            self.memory.put(booking)

    def test_list(self):
        """Retrieve all items from one table."""
        bookings = self.memory.get("transaction")
        self.assertEqual(len(bookings), 10)
        for i, j in enumerate(bookings):
            self.assertEqual(j.amount, 50 + i)
            self.assertEqual(j.description, f"list transaction {i}")

    def test_operators(self):
        """Verify that comparison operators can be used."""
        today = datetime.datetime.now()
        bookings = self.memory.get(*(self.memory.transaction.timestamp <= today,))
        self.assertEqual(len(bookings), 10)
        for i in bookings:
            self.assertTrue(i.timestamp <= today)

    def test_missing_table(self):
        """Operators with missing table should return None."""
        self.memory.get(self.memory.nonexisting.timestamp >= False)


@b.add_memory()
class DynamicFields(b.TestCase):
    """Create memory structures with dynamic field generation."""

    def test(self):
        """Dynamic field must generate id."""
        booking = Transaction(50, "payment for buffer")
        self.memory.put(booking)
        new_booking = self.memory.get.transaction()
        self.assertEqual(booking.id, new_booking.id)
        self.assertEqual(booking.timestamp, new_booking.timestamp)
        self.memory.put(booking)

    def test_wrong_input(self):
        """Dynamic field with wrong input."""

        @dataclass
        class WrongDynamic:
            def add_id(self):
                return self

        with self.assertRaises(membank.GeneralMemoryError):
            self.memory.put(WrongDynamic)


@b.add_memory()
class CreateRead(b.TestCase):
    """Create simple memory structure, add item and get it back."""

    def assert_equal(self, item1, item2):
        """Assert two dogs are equal."""
        for i in ["breed", "color", "weight"]:
            self.assertEqual(getattr(item1, i), getattr(item2, i))
        self.assertIn(str(item1)[:-1], str(item2))

    def test(self):
        """Read and create memory."""
        dog = Dog("Puli")
        self.memory.put(dog)
        self.memory.put(dog)  # puts are idempotent
        new_dog = self.memory.get.dog()
        self.assert_equal(dog, new_dog)
        self.memory.put(new_dog)  # one can put the got thing back
        self.assert_equal(dog, new_dog)

    def test_file_path_absolute(self):
        """Create sqlite with file path."""
        memory = membank.LoadMemory(b.DBPath.ABSOLUTE)
        memory.reset()
        old_dog = Dog("red")
        memory.put(old_dog)
        memory.put(old_dog)
        new_dog = memory.get.dog()
        for i in ["breed", "color", "weight"]:
            self.assertEqual(getattr(old_dog, i), getattr(new_dog, i))

    def test_file_path_relative(self):
        """Create sqlite with relative file path."""
        memory = membank.LoadMemory(b.DBPath.RELATIVE)
        self.assertTrue(memory)


class UpdateHandling(b.TestCase):
    """Do update existing field."""

    def test(self):
        """Create and update."""
        memory = membank.LoadMemory()
        memory.put(Transaction(6.5, "Small post to update"))
        booking = memory.get.transaction()
        self.assertEqual(booking.amount, 6.5)
        booking.amount = 6.6
        memory.put(booking)
        booking = memory.get.transaction()
        self.assertEqual(booking.amount, 6.6)


class UpdateWithDBChange(b.TestCase):
    """Do update with database schema changes."""

    def setUp(self):
        """Add memory and setup initial test data."""
        super().setUp()
        self.memory = membank.LoadMemory()
        self.dog = UpdatedDog("Puli")
        self.memory.put(self.dog)

    def tearDown(self):
        """Restore the dog to original state."""
        super().tearDown()
        global UpdatedDog

        @dataclass
        class UpdatedDog:
            """Version 2 for Dog."""

            breed: str
            color: str = "black"

    def test_update_changed(self):
        """Update item that is changed."""
        # pylint: disable=global-statement
        self.assertEqual(self.dog, self.memory.get.updateddog())
        global UpdatedDog

        @dataclass
        class UpdatedDog:
            """Version 2 for Dog."""

            breed: str
            new_field: str
            color: str = "black"
            weight: float = 0.0

        old_dog = self.memory.get.updateddog()
        self.assertEqual(self.dog.breed, old_dog.breed)
        next_dog = self.memory.get("updateddog")
        self.assertTrue(len(next_dog), 1)
        self.assertEqual(self.dog.breed, next_dog[0].breed)
        dog = UpdatedDog("NewPuli", "something")
        self.memory.put(dog)
        new_dog = self.memory.get.updateddog(breed="NewPuli")
        self.assertEqual(dog, new_dog)

    def test_changed_read(self):
        """Should be possible to read changed item."""
        global UpdatedDog

        @dataclass
        class UpdatedDog:
            """Version 2 for Dog."""

            breed: str
            new_field: str
            color: str = "black"
            weight: float = 0.0

        dog = self.memory.get("updateddog")
        self.assertTrue(len(dog), 1)
        dog = dog[0]
        next_dog = self.memory.get.updateddog()
        self.assertEqual(next_dog, dog)
        self.assertEqual(self.dog.breed, dog.breed)
        self.assertEqual(self.dog.color, dog.color)
        self.assertEqual("Puli", dog.breed)
        self.assertEqual("black", dog.color)
        self.assertEqual(None, dog.new_field)


class UpdateWithKey(b.TestCase):
    """Do update with database schema changes."""

    def test_update_changed_with_key(self):
        """Update item that is changed and has a key field."""
        global Cat
        cat = Cat("Ronalo")
        memory = membank.LoadMemory()
        memory.put(cat)

        @dataclass
        class Cat:
            """Version 2 with key field."""

            id: str = dataclasses.field(default=None, metadata={"key": True})
            breed: str = "unknown"

        cat = Cat("Romber")
        memory.put(cat)
        new_cat = memory.get.cat(id="Romber")
        self.assertEqual(cat, new_cat)
        self.assertEqual(new_cat.breed, "unknown")


class LoadMemoryErrorHandling(b.TestCase):
    """Handle errors on LoadMemory init."""

    def test_wrong_scheme(self):
        """Unrecognised scheme should fail."""
        with self.assertRaises(membank.errors.GeneralMemoryError):
            membank.LoadMemory(url="jumbo://www.zoozl.net")

    def test_wrong_path(self):
        """Invalid paths should fail."""
        with self.assertRaises(membank.errors.GeneralMemoryError):
            membank.LoadMemory(url="berkeleydb://:memory:")
        with self.assertRaises(membank.errors.GeneralMemoryError):
            membank.LoadMemory(url="sqlite://www.zoozl.net/gibberish")
        with self.assertRaises(membank.errors.GeneralMemoryError):
            membank.LoadMemory(url=dict(id="path"))


@b.add_memory()
class PutMemoryErrorHandling(b.TestCase):
    """Handle errors on LoadMemory.put function."""

    def test_wrong_input(self):
        """Input should fail if not namedtuple instance."""
        with self.assertRaises(membank.errors.GeneralMemoryError):
            self.memory.put("blblbl")

        @dataclass
        class UnsupportedType:
            done: Dog

        with self.assertRaises(membank.errors.GeneralMemoryError):
            self.memory.put(UnsupportedType)
        with self.assertRaises(membank.errors.GeneralMemoryError):
            self.memory.put(Dog)
        with self.assertRaises(membank.errors.GeneralMemoryError):
            self.memory.put(Dog(1))

    def test_reserved_name(self):
        """Input should fail if reserved name."""

        @dataclass
        class __meta_dataclasses__:
            id: str

        with self.assertRaises(membank.errors.GeneralMemoryError):
            self.memory.put(__meta_dataclasses__("ad"))

        @dataclass
        class Put:
            id: str

        with self.assertRaises(membank.errors.GeneralMemoryError):
            self.memory.put(Put("ad"))


class GetMemoryErrorHandling(b.TestCase):
    """Handle errors on LoadMemory.get function."""

    def test_none_existing_table(self):
        """Return None if not existing table."""
        memory = membank.LoadMemory(b.DBPath.RELATIVE)
        self.assertIsNone(memory.get.thisdoesnotexist())
        self.assertEqual(memory.get("thisdoesnotexist"), [])
        self.assertEqual(memory.get("thisdoesnotexist", relation=5), [])

    def test_attribute_error(self):
        """Fetching non existing attribute should fail."""
        memory = membank.LoadMemory()
        memory.put(Dog("lol"))
        with self.assertRaises(membank.errors.GeneralMemoryError) as error:
            memory.get(memory.dog.super_breed == "lol")
        self.assertIn("does not hold", str(error.exception))
        with self.assertRaises(membank.errors.GeneralMemoryError) as error:
            memory.get(breed="lol")


@b.add_memory()
class GetWithKeywords(b.TestCase):
    """Get objects by keywords."""

    def test(self):
        """Simple get with keyword."""
        cat = Cat("1", "Ronalo")
        self.memory.put(cat)
        db_cat = self.memory.get.cat(name="Ronalo")
        self.assertEqual(cat, db_cat)
        # exactly the same call still works the same
        db_cat = self.memory.get.cat(name="Ronalo")
        self.assertEqual(cat, db_cat)
        db_cat = self.memory.get("cat", name="Ronalo")
        self.assertTrue(len(db_cat), 1)
        db_cat = db_cat[0]
        self.assertEqual(cat, db_cat)


@b.add_memory()
class GetComplexObject(b.TestCase):
    """Get nested objects."""

    def test(self):
        """Create nested object and retrieve it."""
        dog1, dog2 = Dog("Bulldog"), Dog("Poodle")
        team = DogTeam("Pulis", [dog1, dog2])
        self.memory.put(team)
        db_team = self.memory.get.dogteam(name="Pulis")
        self.assertEqual(team.name, db_team.name)
        self.assertEqual(team.team[0], db_team.team[0])
        self.assertEqual(team, db_team)
        db_team = self.memory.get("dogteam", name="Pulis")
        self.assertTrue(len(db_team), 1)
        db_team = db_team[0]
        self.assertEqual(team, db_team)



"""
Defines interfaces exposed to library user.

>>> from dataclasses import dataclass, field
>>> from membank import LoadMemory
>>> memory = LoadMemory()
>>> @dataclass
... class User:
...     id: int = field(metadata={"key": True})
...     name: str
>>> memory.put(User(id=1, name="Alice"))
"""

from membank.errors import (GeneralMemoryError, MemoryFilteringError,
                            MemoryOutOfSyncError, MemoryTableDoesNotExist)
from membank.interface import LoadMemory

__all__ = [
    "LoadMemory",
    "GeneralMemoryError",
    "MemoryTableDoesNotExist",
    "MemoryFilteringError",
    "MemoryOutOfSyncError",
]



"""Common methods and constants for all tests."""

import enum
import functools
import os
import unittest

import membank


TEST_DATABASE_PATH = "tests/data/test_database.db"


class DBPath(str, enum.Enum):
    """Enum for database paths."""

    RELATIVE = f"sqlite://{TEST_DATABASE_PATH}"
    ABSOLUTE = f"sqlite://{os.getcwd()}/{TEST_DATABASE_PATH}"


class TestCase(unittest.TestCase):
    """Generic TestCase class."""


def add_memory(path=None, reset=False):
    """Decorate TestCase with memory."""
    if path is not None and not isinstance(path, str):
        raise TypeError(f"path must be a string, was {path}")

    def decorator(cls):
        """Add memory to TestCase cls."""

        @functools.wraps(cls, updated=())
        class MemoryTestCase(cls):
            """Add memory to TestCase."""

            def setUp(self):
                """Set up memory."""
                if path:
                    self.memory = membank.LoadMemory(path)
                else:
                    self.memory = membank.LoadMemory()
                super().setUp()

            def tearDown(self):
                """Tear down memory."""
                super().tearDown()
                if reset:
                    self.memory.reset()

        return MemoryTestCase

    return decorator



"""Functions to interact with database."""
import dataclasses
import datetime

from alembic.migration import MigrationContext
from alembic.operations import Operations
import sqlalchemy as sa

from membank import errors as e


# Mapping of Python types with SQL types
SQL_TABLE_TYPES = {
    float: sa.Float,
    str: sa.String,
    int: sa.Integer,
    datetime.datetime: sa.DateTime,
    datetime.date: sa.Date,
    bytes: sa.LargeBinary,
    bool: sa.Boolean,
    dict: sa.JSON,
    list: sa.JSON,
}


def get_sql_col_type(py_type):
    """From Python data type py_type returns SQL type."""
    if py_type in SQL_TABLE_TYPES:
        return SQL_TABLE_TYPES[py_type]
    raise e.GeneralMemoryError(f"Type {py_type} is not supported")


def make_stmt(sql_table, return_class, *filtering, **matching):
    """Do select stmt."""
    stmt = sa.select(*_get_selectables(sql_table, return_class))
    return filter_stmt(stmt, sql_table, *filtering, **matching)


def filter_stmt(stmt, sql_table, *filtering, **matching):
    """Prepare SQL statement and returns it."""
    if matching:
        for key, value in matching.items():
            stmt = stmt.where(getattr(sql_table.c, key) == value)
    if filtering:
        for item in filtering:
            stmt = stmt.where(item)
    return stmt


def get_item(sql_table, engine, return_class, **matching):
    """Get item from table."""
    stmt = make_stmt(sql_table, return_class, **matching)
    with engine.connect() as conn:
        cursor = execute_stmt(conn, stmt).first()
    return return_class(*cursor) if cursor else None


def get_list(sql_table, engine, return_class, *filtering, **matching):
    """Get list of items from table."""
    stmt = make_stmt(sql_table, return_class, *filtering, **matching)
    return get_from_sql(return_class, stmt, engine)


def delete_item(sql_table, engine, **matching):
    """Execute delete stmt."""
    stmt = sa.delete(sql_table)
    stmt = filter_stmt(stmt, sql_table, **matching)
    with engine.connect() as conn:
        execute_stmt(conn, stmt)
        conn.commit()


def get_from_sql(return_class, stmt, engine):
    """Get all items from table as per SQL statement."""
    with engine.connect() as conn:
        cursor = execute_stmt(conn, stmt)
        return [return_class(*i) for i in cursor]


def update_item(sql_table, engine, item, key=None):
    """Create or update an item in table."""
    stmt = sa.select(sql_table)
    if key:
        col = _getmemattr(sql_table.c, key)
        val = getattr(item, key)
        stmt = stmt.where(col == val)
    else:
        for col, val in _unpack_values(item, sql_table):
            stmt = stmt.where(col == val)
    with engine.connect() as conn:
        rows = execute_stmt(conn, stmt)
        record = rows.first()
    if not record or (record and key):
        if record and key:
            col = getattr(sql_table.c, key)
            val = getattr(item, key)
            stmt = sql_table.update()
            stmt = stmt.where(col == val)
        else:
            stmt = sql_table.insert()
        stmt = stmt.values(dict(_unpack_values(item, sql_table)))
        with engine.connect() as conn:
            with conn.begin():
                execute_stmt(conn, stmt)


def execute_stmt(conn, stmt):
    """Execute SQL statement."""
    try:
        return conn.execute(stmt)
    except sa.exc.StatementError as error:
        msg = str(error.orig)
        if isinstance(error.orig, TypeError):
            msg += " Invalid field type, possibly use 'encode' in metadata"
        if "no such table:" in msg:
            msg = msg.split(":", maxsplit=1)[-1].strip()
            raise e.MemoryTableDoesNotExist(msg) from None
        raise e.GeneralMemoryError(msg) from None


def sync_table(sql_table, engine, obj):
    """Sync table as per obj values."""
    with engine.connect() as conn:
        alembic = Operations(MigrationContext.configure(conn))
        fields = dataclasses.fields(obj)
        for field in fields:
            if field.name not in sql_table.c:
                col_type = get_sql_col_type(field.type)
                col = sa.Column(field.name, col_type)
                alembic.add_column(sql_table.name, col)


def introspect_table_fields(sql_table):
    """Introspect table fields."""
    fields = []
    for col in sql_table.c:
        col_type = None
        for py_type, sql_type in SQL_TABLE_TYPES.items():
            if isinstance(col.type, sql_type):
                col_type = py_type
                break
        if not col_type:
            msg = f"Column '{col.name}' has unsupported type '{col.type}'"
            raise e.GeneralMemoryError(msg)
        fields.append((col.name, col_type))
    return fields


def create_table(table, instance, engine):
    """Add a memory attribute.

    Memory attribute must be instance of dataclass In database words
    this adds a new Table
    """
    with engine.connect() as conn:
        alembic = Operations(MigrationContext.configure(conn))
        fields = dataclasses.fields(instance)
        cols = []
        for field in fields:
            col_type = get_sql_col_type(field.type)
            col = sa.Column(field.name, col_type)
            cols.append(col)
        try:
            alembic.create_table(table, *cols)
        except sa.exc.OperationalError as error:
            msg = error.args[0]
            if "table" in msg and "already exists" in msg:
                msg = f"Table {table} already exists. Use change instead"
                raise e.GeneralMemoryError(msg) from None


class FilterOperator():
    """Allows to filter memory items by expressions."""

    def __init__(self, name, meta):
        """Initialize."""
        self.__name = name
        if name in meta.tables:
            self.__sql_table = meta.tables[name]
        else:
            self.__sql_table = None
        self.__column = False
        self.__operator = False

    def __lt__(self, other):
        """Operations with <."""
        op = self.__column < other if self.__operator else None
        return self.__sql_table, op

    def __le__(self, other):
        """Operations with <=."""
        op = self.__column <= other if self.__operator else None
        return self.__sql_table, op

    def __eq__(self, other):
        """Operations with ==."""
        op = self.__column == other if self.__operator else None
        return self.__sql_table, op

    def __ne__(self, other):
        """Operations with !=."""
        op = self.__column != other if self.__operator else None
        return self.__sql_table, op

    def __gt__(self, other):
        """Operations with >."""
        op = self.__column > other if self.__operator else None
        return self.__sql_table, op

    def __ge__(self, other):
        """Operations with >=."""
        op = self.__column >= other if self.__operator else None
        return self.__sql_table, op

    def __getattr__(self, name):
        if getattr(self.__sql_table, "name", False):
            self.__column = getattr(self.__sql_table.c, name, False)
            if self.__column is False:
                msg = f"'{self.__name}' does not hold '{name}'"
                raise e.GeneralMemoryError(msg)
            self.__operator = True
        return self


def _getmemattr(obj, name):
    """Get memory attribute."""
    try:
        return getattr(obj, name)
    except AttributeError:
        raise e.MemoryOutOfSyncError(obj, name) from None


def _unpack_values(obj, table):
    """Unpack values from object using table as reference."""
    for i in dataclasses.fields(obj):
        key = _getmemattr(table.c, i.name)
        val = getattr(obj, i.name)
        if "encode" in i.metadata:
            val = i.metadata["encode"](val)
        yield key, val


def _get_selectables(table, obj):
    """Get column names in their respective order as per dataclass object."""
    for i in dataclasses.fields(obj):
        key = _getmemattr(table.c, i.name)
        yield key
